using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using System.IO.Ports;
using System.Windows.Forms;

namespace XCD_Communication {

    public class Communication {

        public enum Command : byte {
            Move = 1,
            Assign = 3,
            Home = 4,
            VelocityLoop = 6,
            OpenLoop = 7,
            Nop = 10,
            Save = 13,
            Restore = 14,
            SetSerial = 15,
            SetAddress = 16,
            Enable = 17,
            Disable = 18,
            Info = 19,
            Monitor = 20,
            MonitorAddr = 21,
            Mode = 22,
            Kill = 23,
            Report = 26,
            FeedbackDirection = 30,
            OutDirection = 31,
            Config = 32,
        };

        public struct CommandDescriptor {
            public string ComTxt;
            public int Timeout;
            public Int16 nParams;
            public byte[] ParamType;
            public CommandDescriptor(string aComTxt, int aTimeout, Int16 anParams, byte[] aParamType) {
                ComTxt = aComTxt; Timeout = aTimeout; nParams = anParams; ParamType = aParamType;
            }
        };

        public const byte PT_I8 = 0, PT_I16 = 1, PT_I32 = 2, PT_F32 = 3;     // Parameter types int8/16/32/float

        public CommandDescriptor[] CommmandTable = {
            new CommandDescriptor("invalid", 0, 10, new byte[] {PT_F32,PT_F32,PT_F32,PT_F32,PT_F32,PT_F32,PT_F32,PT_F32,PT_F32,PT_F32}),                                //0
            new CommandDescriptor("move", 0, 1, new byte[] {PT_F32}),                    //1
            new CommandDescriptor("invalid", 0, 0, null),                                //2
            new CommandDescriptor("assign", 0, 14, new byte[] {PT_I16, PT_F32, PT_I16, PT_F32, PT_I16, PT_F32, PT_I16, PT_F32, PT_I16, PT_F32, PT_I16, PT_F32, PT_I16, PT_F32}),                                //3
            new CommandDescriptor("home", 0, 4, new byte[] {PT_I8, PT_F32, PT_F32, PT_F32}),     //4
            new CommandDescriptor("invalid", 0, 0, null),                                //5
            new CommandDescriptor("velocityloop", 0, 1, new byte[] {PT_F32}),            //6
            new CommandDescriptor("openloop", 0, 1, new byte[] {PT_F32}),                //7
            new CommandDescriptor("invalid", 0, 0, null),                                //8
            new CommandDescriptor("invalid", 0, 0, null),                                //9
            new CommandDescriptor("nop", 0, 0, null),                                    //10
            new CommandDescriptor("invalid", 0, 0, null),                                //11
            new CommandDescriptor("invalid", 0, 0, null),                                //12
            new CommandDescriptor("save", 5000, 2, new byte[] {PT_I8, PT_I8}),           //13
            new CommandDescriptor("invalid", 0, 0, null),                                //14
            new CommandDescriptor("setserial", 0, 3, new byte[] {PT_I8, PT_I8, PT_I32}), //15
            new CommandDescriptor("setaddress", 0, 3, new byte[] {PT_I8, PT_I8, PT_I8}), //16
            new CommandDescriptor("enable", 0, 0, null),                                //17
            new CommandDescriptor("disable", 0, 0, null),                                //18
            new CommandDescriptor("invalid", 0, 0, null),                                //19
            new CommandDescriptor("monitor", 0, 3, new byte[] {PT_I8, PT_I16, PT_F32}),   //20
            new CommandDescriptor("monitoraddr", 0, 3, new byte[] {PT_I8, PT_I16, PT_F32}), //21
            new CommandDescriptor("pwmmode", 0, 3, new byte[] {PT_I8, PT_I8, PT_I8}),                  //22
            new CommandDescriptor("kill", 0, 0, null),                                //23
            new CommandDescriptor("invalid", 0, 0, null),                                //24
            new CommandDescriptor("invalid", 0, 0, null),                                //25
            new CommandDescriptor("report", 0, 10, new byte[] {PT_I16, PT_I16, PT_I16, PT_I16, PT_I16, PT_I16, PT_I16, PT_I16, PT_I16, PT_I16}), //26
            new CommandDescriptor("invalid", 0, 0, null),                                //27
            new CommandDescriptor("invalid", 0, 0, null),                                //28
            new CommandDescriptor("invalid", 0, 0, null),                                //29
            new CommandDescriptor("feedbackdirection", 0, 1, new byte[] {PT_I8}),        //30
            new CommandDescriptor("outdirection", 0, 1, new byte[] {PT_I8}),             //31
            new CommandDescriptor("config", 0, 8, new byte[] {PT_I16, PT_F32, PT_F32, PT_F32, PT_F32, PT_F32, PT_F32, PT_F32}),  //32
        };

        // Synchronization constants for RS232 communication
        const byte RS_SYNC1 = 0xe4;
        const byte RS_SYNC2 = 0xa5;

        const byte VARREAD_CHUNK = 10;      // Maxinum number of reported vars in one command
        const byte VARWRITE_CHUNK = 7;      // Maxinum number of assigned vars in one command

        public string PortName = "COM1";    // Name of serial port
        public static SerialPort Port;      // SerialPort object

        public byte DestinationAddress = 0;            // Destination address

        /// <summary> Open port using PortName </summary>
        public bool OpenPort() {
            if (!(Port == null)) {
                Port.Close();
                //Reading.Join()
            }
            try {
                Port = new SerialPort(PortName, 115200);
                Port.DataBits = 8;
                Port.DiscardNull = false;
                Port.DtrEnable = false;
                Port.Handshake = Handshake.None;
                Port.Parity = Parity.None;
                Port.ReceivedBytesThreshold = 1;
                Port.StopBits = StopBits.One;
                Port.ReadBufferSize = 4096;
                Port.WriteBufferSize = 2048;
                Port.ReadTimeout = 1000;
                Port.WriteTimeout = 1000;
                Port.Open();
                Port.DiscardInBuffer();
            }
            catch {
                Port = null;
                throw new InvalidOperationException("Cannot open serial port");
            }
            return true;
        }

        /// <summary> Validate the port is open, reopen if necessary </summary>
        public void ValidatePort() {
            if (Port == null) {
                OpenPort();
            }
        }

        /// <summary> Read input byte stream, find two synchro bytes RS_SYNC1/RS_SYNC2 </summary>
        private void Synchronize() {
            byte b1;
            do {
                b1 = (byte)Port.ReadByte();
                while (b1 == RS_SYNC1) {
                    b1 = (byte)Port.ReadByte();
                    if (b1 == RS_SYNC2) return;
                }
            }
            while (true);
        }

        /// <summary> Send command to communication port </summary>
        private void WriteCommand(byte[] command, byte count) {
            ValidatePort();
            byte[] Sync = { RS_SYNC1, RS_SYNC2, DestinationAddress, count };    // UART prefix
            Port.Write(Sync, 0, 4);             // Send prefix
            Port.Write(command, 0, count);      // Send body
        }

        /// <summary> Read controller reply </summary>
        public byte[] ReadReply() {
            ValidatePort();
            for (int Repeat = 0; Repeat < 1; Repeat++) {        // Retry in case of communication error
                try {
                    Synchronize();                      // Find RS_SYNC1/RS_SYNC2
                    Port.ReadByte();                    // Controller address
                    byte Count = (byte)Port.ReadByte(); // Reply body length
                    byte[] R = new byte[Count]; 
                    for (byte i = 0; i < Count; i++) {
                        R[i] = (byte)Port.ReadByte();   // Read reply body
                    }
                    return R;
                }
                catch {
                    //                return null;
                }
            }
            // We are here only if two attempts failed
            throw new InvalidOperationException("Communication failure");
        }

        /// <summary> Send command and read reply </summary>
        public byte[] ExecuteCommand(byte[] Command, byte Count) {
            lock (this) {           // To ensure multithread access
                ValidatePort();
                Port.DiscardInBuffer();
                WriteCommand(Command, Count);
                System.Threading.Thread.Sleep(3);
                return ReadReply();
            }
        }

        /// <summary> Build command, parse the parameters using parTypes, 
        ///   send command and read reply
        /// </summary>
        public byte[] ExecuteCommand(byte command, double[] pars, byte[] parTypes) {
            byte[] Message = new byte[20];
            Message[0] = command;
            int j = 1;
            if (pars != null) {
                for (int i = 0; i < pars.Length; i++) {
                    if ((parTypes == null) || (i >= parTypes.Length)) {
                        throw new InvalidOperationException("More parameters than required");
                    }
                    if (parTypes[i] == PT_I8) {     // Int8
                        Message[j++] = (byte)pars[i];
                    } else if (parTypes[i] == PT_I16) {     // Int16
                        Array.Copy(BitConverter.GetBytes((Int16)pars[i]), 0, Message, j, 2);
                        j += 2;
                    } else if (parTypes[i] == PT_I32) {     // Int32
                        Array.Copy(BitConverter.GetBytes((Int32)pars[i]), 0, Message, j, 4);
                        j += 4;
                    } else if (parTypes[i] == PT_F32) {     // float                
                        Array.Copy(BitConverter.GetBytes((float)pars[i]), 0, Message, j, 4);
                        j += 4;
                    }
                }
            }
            int SaveTimeout = Port.ReadTimeout;
            if (CommmandTable[command].Timeout != 0) {      // For long commands, e.g. SAVE
                Port.ReadTimeout = CommmandTable[command].Timeout;
                Application.UseWaitCursor = true;
            }
            try {
                return ExecuteCommand(Message, (byte)j);
            }
            finally {
                Port.ReadTimeout = SaveTimeout;
                Application.UseWaitCursor = false;
            }
        }

        /// <summary> Build command, parse the parameters using CommandTable, 
        ///   send command and read reply
        /// </summary>
        public byte[] ExecuteCommand(byte command, double[] parameters, bool DisplayErrorMessages) {
            try {
                // For unknown command, use default parmater parsing
                byte pars = (CommmandTable[command].ComTxt == "invalid") ? (byte)0 : command;
                byte[] Reply = ExecuteCommand(command, parameters, CommmandTable[pars].ParamType);
                if (Reply == null) return null;
                if (Reply[1] == 2) throw new InvalidOperationException("Command execution failed");
                return Reply;
            }
            catch (Exception ex) {
                if (DisplayErrorMessages && !Thread.CurrentThread.IsBackground) {
                    MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
                return null;
            }
        }

        /// <summary> Helper function for commands with a single real parameter </summary>
        public byte[] ExecuteCommand(byte command, double parameter, bool DisplayErrorMessages) {
            return ExecuteCommand(command, new double[] { parameter }, DisplayErrorMessages);
        }

        /// <summary> Read variable </summary>
        public float ReadVar(short VarId) {
            byte[] Rep = ExecuteCommand((byte)Command.Report, new double[] { VarId }, false);
            if (Rep == null)
                throw new InvalidOperationException("Operation failed");
            if (Rep.Length < 6)
                throw new InvalidOperationException("Operation failed");
            return BitConverter.ToSingle(Rep, 2);
        }

        /// <summary> Write variable </summary>
        public void WriteVar(short VarId, float Val) {
            byte[] Rep = ExecuteCommand((byte)Command.Assign, new double[] { VarId, Val }, true);
        }
    }
}


